/* fCPM: portable CPM emulator
 *
 * This file contains CPM specific initialization.
 */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
/*
#include <sys/ioctl.h>
*/
#include <signal.h>
#include "CPM.h"
#include "bios.h"

#define WRALL		(0)
#define WRDIR		(1)
#define WRUAL		(2)
#define WREAD		(3)

/* disk file names for CP/M disks */
char name[NDISKS][50];

FILE *fppun = NULL;
FILE *fprdr = NULL;
FILE *fplist = NULL;

/* getopt stuff */
extern char *optarg;
extern int optind, opterr;

char bootfile[50];

byte Verbose = 1;

void goodbye();
char *malloc();

unsigned char *RAM;
char *dmaptr;

/* assumes 32 bit architecture! */
/* disk variables */
union {
	int track;
	unsigned char trck[4];
} trackno;
int sector;

main(int argc,char *argv[])
{
	int ret;
	int ret2;
	int i;
	int c;

	/* load default boot image in case user doesn't specify */
	strcpy(bootfile, BOOT_SYS);

	/* load default drive name in case user doesn't specify */
	strcpy(name[0], DISK_SYS);

	/* check arguments (if any) */

	opterr = 1;	/* Make sure it complains about illegal options.
				 */

	while ((c = getopt(argc,argv,"a:b:c:d:e:f:g:h:i:")) != -1) {
		switch(c) {
		case 'a':
			/* drive A */
			strcpy(name[0], optarg);
			break;
		case 'b':
			/* drive B */
			strcpy(name[1], optarg);
			break;
		case 'c':
			/* drive C */
			strcpy(name[2], optarg);
			break;
		case 'd':
			/* drive D */
			strcpy(name[3], optarg);
			break;
		case 'e':
			/* drive E */
			strcpy(name[4], optarg);
			break;
		case 'f':
			/* drive F */
			strcpy(name[5], optarg);
			break;
		case 'g':
			/* drive G */
			strcpy(name[6], optarg);
			break;
		case 'h':
			/* drive H */
			strcpy(name[7], optarg);
			break;
		case 'i':
			/* Override Boot Image */
			strcpy(bootfile, optarg);
			break;
		default:
			/* give quickie usage message */
			fprintf(stderr, "Usage: %s [-a filename] through [-h filename]\n", argv[0]);
			fprintf(stderr, "The default filename for drive A is drivea.\n");
			fprintf(stderr, "Also [-i bootimage] will override default boot image of e400sys.img\n");
			fflush(stderr);
			exit(1);
		}
	}

	/* open rdr, pun and list devices */
	if ((fplist = fopen("printer.cpm","ab")) == NULL) {
		fprintf(stderr, "Cannot open printer.cpm\nAborting...\n");
		exit(1);
	}
	if ((fppun = fopen("punch.cpm", "ab")) == NULL) {
		fprintf(stderr, "Cannot open punch.cpm\nAborting...\n");
		exit(1);
	}
	/* rdr input - not an error if nothing is there
	 * If it opens, OK, if not that's OK, too */
	fprdr = fopen("reader.cpm", "rb");

	/* trap signals so the terminal doesn't go to hell 
	 * when we exit
	 */
	signal(SIGINT, goodbye);

	/* initialize drives */
	ret2 = x_init();

	/* major problems */
	if (ret2 == 1) {
		fprintf(stderr, "x_init failed.\r\n");
		goodbye(99);
	}

	/* kick off CP/M */
	StartCPM();

	/* Return memory */
	TrashCPM();

	/* we should exit through the goodbye() function
	 * to restore terminal modes and close disk drives
	 */
	goodbye(99);	

	/* The logic will never get here */
	return(0);
}


/* malloc memory, initialize page 0, and put bootfile (default BOOT_SYS)
 * into virtual memory at e400H.  Fid must contain a complete CPM image of
 * CCP, BDOS and BIOS.  Then call Z80 engine to do its thing.
 */
StartCPM()
{
	word A;
	reg R;
	int *T,I,J,K;
	byte *P;
    FILE *fp;
	int address = 0xe400;
    int c;
	int i = 0;

	/*** STARTUP CODE starts here: ***/

	T=(int *)"\01\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
#ifdef LSB_FIRST
	if (*T!=1) {
		fprintf(stderr, "********** This machine is high-endian. **********\r\n");
		fprintf(stderr, "Take #define LSB_FIRST out and compile CPM again.\r\n");
		fflush(stderr);
		goodbye(99);
	}
#else
	if (*T==1) {
		fprintf(stderr, "********* This machine is low-endian. **********\r\n");
		fprintf(stderr, "Insert #define LSB_FIRST and compile fMSX again.\r\n");
		goodbye(99);
	}
#endif

	/* Misc notes from tlb
	 *
	 * In standard e400sys.img file, The CCP is assembled to run
	 * at E400H, CPM 2.2 BDOS runs at EC00H and the BIOS
	 * jump table is at page FA00H.  The BIOS ends at FBFFH.
	 */

	if(Verbose) {
		printf("Allocating 64kB for address space...\r\n");
	}
	if(!(RAM=malloc(4*PAGESIZE))) {
		fprintf(stderr,"RAM malloc failed!\r\n");
		fflush(stderr);
		goodbye(99);
	}
	
	/* clear all virtual memory
	 */
    memset(RAM,ZERORAM,4*PAGESIZE);
	/* poke desired values into virtual memory (initialize page 0)
	 * for virtual CP/M machine.
	 */

	/* warm boot at location 0000H */
	RAM[0] = 0xc3;	/* warm bios jump opcode */
	RAM[1] = 0x03;	/* low byte of address - 2nd entry in bios jump table */
	RAM[2] = 0xfa;	/* high byte of 64k bios base page */

	/* iobyte at location 0003H */
	RAM[3] = 0x8;

	/* current disk at location 0004H */
	RAM[4] = 0x0;

	/* BDOS entry at location 0005H */
	RAM[5] = 0xc3;	/* BDOS entry jump opcode */
	RAM[6] = 0x06;	/* low byte of address - first 6 chars of bdos is ser no */
	RAM[7] = 0xec;	/* high byte of EC00H BDOS */

	/* open file */
    if ((fp = fopen(bootfile,"rb")) == NULL) {
		fprintf(stderr,"Open of %s failed!  Bye!\r\n", BOOT_SYS);
		fflush(stderr);
		goodbye(99);
	}
	/* load OS file data into Z80 memory */
	while ((c = getc(fp)) != EOF) {
		RAM[address] = c;
		address++;
	}

	fclose(fp);

	/* sanity check */
	if (address < 0xfb00) {
		fprintf(stderr,"Invalid load, address = %04X\r\n",address);
		fflush(stderr);
		goodbye(99);
	}

	/* set Program Counter to 0000H */
	R.PC.W = 0x0000;

	if(Verbose) printf("OK\r\nRUNNING CPM CODE...\r\n");
	/* run CPU */
	A=Z80(RAM,R);

	if(Verbose) printf("EXITED at PC = %Xh.\r\n",A);
	return(1);

}


/****************************************************************/
/*** Emulate BIOS calls. This function is called on the ED FE ***/
/*** instruction to emulate disk/tape access, etc.            ***/
/****************************************************************/
/* The BIOS number is located at the next byte in memory.
 * Switch on the BIOS function and go.
 */
void Patch(reg *R)
{
	int c;

    switch(RAM[R->PC.W]) {
		case 0x00:	/* cold boot */
			R->AF.B.h = 0x00;
			break;

		case 0x01:	/* warm boot */
			R->AF.B.h = i_boot() ? 0xFF : 0x00;
			R->BC.B.l = R->AF.B.h;
			break;

		case 0x02:	/* console status */
			R->AF.B.h = x_const() ? 0xFF : 0x00;
			break;

		case 0x03:	/* console input */
			c = x_conin() & 0x7F;
			if (c == 10) {	/* do carriage return from new line xlate */
				R->AF.B.h = 13;
			}
			else {
				R->AF.B.h = c;
			}
			break;

		case 0x04:	/* console output */
			x_conot(R->BC.B.l & 0x7F);
			break;

		case 0x05:	/* list output */
			x_lstot(R->BC.B.l & 0x7F);
			break;

		case 0x06:	/* punch device output */
			x_punot(R->BC.B.l);
			break;

		case 0x07:	/* reader device input */
			R->AF.B.h = x_rdrin();
			break;

		case 0x0A:	/* set track */
			/* values are in BC register pair */
#ifdef LSB_FIRST
			/* LITTLE ENDIAN */
			trackno.trck[3] = '\0';
			trackno.trck[2] = '\0';			
			trackno.trck[1] = R->BC.B.h;	/* high byte */
			trackno.trck[0] = R->BC.B.l;	/* low byte */
#else
			/* BIG ENDIAN */
			trackno.trck[0] = '\0';
			trackno.trck[1] = '\0';			
			trackno.trck[2] = R->BC.B.h;	/* high byte */
			trackno.trck[3] = R->BC.B.l;	/* low byte */
#endif
			break;
	
		case 0x0B:	/* set sector */
			sector = R->BC.B.l;
			break;

		case 0x0C:	/* set dma */
			dmaptr = &RAM[R->BC.W];
			break;

		case 0x0D:	/* disk read */
			R->AF.B.h = x_rd(dmaptr, RAM[R->DE.W], trackno.track, sector);
			break;

		case 0x0E:	/* disk write */
			R->AF.B.h = x_wr(dmaptr, RAM[R->DE.W], trackno.track, sector);
			break;

		case 0x0F:	/* list status */
			R->AF.B.h = x_lstst() ? 0xFF : 0x00;
			break;

		case 0x10:	/* special BIOS call to leave CP/M */
			/* This stops software Z80 CPU */
			CPURunning = 0;
			break;

		default:
			fprintf(stderr, "Unknown ESC operation: %d\n", RAM[R->PC.W]);
			return;
	}
	/* bump program counter past ESC code */
	R->PC.W++;
	return;
}

/* Boot strap CP/M procedure */

i_boot()
{
	FILE *fp;
	int c;
	int address = 0xe400;

	/* do not clear all virtual memory
	 * as it resets current disk as well as wipes
	 * out any program which may be in the TPA.
	 *
	 */
	/*
    memset(RAM,ZERORAM,4*PAGESIZE);
	*/
	
	/* poke desired values into virtual memory (initialize page 0)
	 */

	/* warm boot at location 0000H */
	RAM[0] = 0xc3;	/* warm bios jump opcode */
	RAM[1] = 0x03;	/* low byte of address - 2nd entry in bios jump table */
	RAM[2] = 0xfa;	/* high byte of 64k bios base page */

	/* BDOS entry at location 0005H */
	RAM[5] = 0xc3;	/* BDOS entry jump opcode */
	RAM[6] = 0x06;	/* low byte of address - first 6 chars of bdos is ser no */
	RAM[7] = 0xec;	/* high byte of EC00H BDOS */

	/* reload CCP/BDOS */

	/* open file */
    if ((fp = fopen(bootfile,"rb")) == NULL) {
		fprintf(stderr,"Open of %s failed!  Bye!\n", bootfile);
		fflush(stderr);
		goodbye(99);
	}

	/* load file data into Z80 memory */
	while ((c = getc(fp)) != EOF) {
		RAM[address] = c;
		address++;
	}

	fclose(fp);

	/* return current working drive */
	return(0);

}

/* default file name interface */
char *
i_sysdsk()
{
    return(DISK_SYS);
}

/****************************************************************/
/*** Write number into given IO port.                         ***/
/****************************************************************/
void DoOut(register byte Port,register byte Value)
{
	/* not needed for generic CP/M environment */
}

/****************************************************************/
/*** Read number from given IO port.                          ***/
/****************************************************************/
byte DoIn(register byte Port)
{
	/* not needed for generic CP/M environment */
}


/****************************************************************/
/*** Refresh screen, check keyboard and sprites. Call this    ***/
/*** function on each interrupt.                              ***/
/****************************************************************/
word Interrupt(void)
{
	/* not needed for generic CP/M environment */
}

/* TrashCPM */
TrashCPM()
{
	/* de allocate memory for Z80 CPU */
	free(RAM);
}


/* signal trap - reset tty modes and leave program
 * 99 is used if the program wants to exit gracefully,
 */
void goodbye(sig)
int sig;
{
	/* close open disks */
	x_exit();

	/* close devices */
	fclose(fplist);
	fclose(fppun);
	if (fprdr != NULL) {
		fclose(fprdr);
	}
	/* bye */
	exit(sig);
}

